/*
    CFS0.11 filesystem implementation for FUSE
    Copyright (C) 2005-2020 Kajtar Zsolt (Soci/Singular) <soci@c64.rulez.org>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
*/

#define FUSE_USE_VERSION 26

#include <fuse.h>
#include <stdio.h>
#include <string.h>
#include <dirent.h>
#include <errno.h>
#include <sys/statfs.h>
#include <stdlib.h>
#include <time.h>
#include <pthread.h>
#include <locale.h>
#include "buffer.h"
#include "block.h"
#include "petscii.h"
#include "file.h"

#ifdef MB_CUR_MAX
#undef MB_CUR_MAX
#endif
#define MB_CUR_MAX 6

enum {
    FTYPE_DEL, FTYPE_PRG, FTYPE_REL, FTYPE_DIR, FTYPE_LNK, FTYPE_MASK = 7
};

static pthread_mutex_t big_lock = PTHREAD_MUTEX_INITIALIZER;
static Offset partition_table; //constant
static time_t epoch_time = 0; //constant
static struct stat stat_default; //constant

typedef struct Partition {
    int valid;
    int bam;
    Petscii name[17];
    Offset rootdir;
    Offset deldir;
    Offset start;
    Offset end;
    off_t size;
    off_t free;
} Partition;

enum { PARTITION_COUNT = 16 };
static Partition partitions[PARTITION_COUNT];

static void offset_insert(unsigned char ide[], Offset ezt) {
    unsigned char old = ide[0] & 0xb0;
    offset_to(ide, ezt);
    ide[0] = (ide[0] & 0x4f) | old;
}

static void fill_in_time(unsigned char *sbuff, time_t ee) {
    struct tm ido2;

    localtime_r(&ee, &ido2);

    ido2.tm_mon++;
    sbuff[0] = (ido2.tm_sec % 60) | ((ido2.tm_mon << 4) & 0xc0);
    sbuff[1] = ido2.tm_min | (ido2.tm_mon << 6);
    sbuff[2] = ((ido2.tm_year - 80) % 100) | ((ido2.tm_hour << 3) & 0xc0);
    sbuff[3] = ido2.tm_mday | (ido2.tm_hour << 5);
}

static const char corr[16] = {
    0, 0, 1, 5, 6, 8, 9, 11, 12, 13, 15, 16, 18, 0, 0, 0
};

static void read_in_time(struct stat *at, const unsigned char *t) {
    static unsigned char last[4];
    static time_t last_time;

    if (memcmp(last, t, 4) != 0) {
        memcpy(last, t, 4);

        if (t[3] & 0x1f) {
            int i = (((t[0] >> 4) & 0x0c) | (t[1] >> 6));
            time_t new_time = t[2] & 0x3f;
            new_time = new_time * 1461 / 4;
            new_time -= (t[2] & 3) == 0 && i < 3;
            new_time += i * 32 - corr[i];
            new_time += t[3] & 0x1f;
            new_time *= 24;
            new_time += ((t[2] >> 3) & 0x18) | (t[3] >> 5);
            new_time *= 60;
            new_time += t[1] & 0x3f;
            new_time *= 60;
            new_time += t[0] & 0x3f;
            last_time = new_time + epoch_time;
        } else {
            last_time = stat_default.st_mtime;
        }
    }
    at->st_mtime = last_time;
    at->st_atime = last_time;
    at->st_ctime = last_time;
}

static int find_free_sector(Partition *partition, Offset *st, int data) {
    Offset itt, fraction, bamm, start = *st;
    unsigned int b;
    Buffer *bam;

    if (start < partition->start || start >= partition->end) start = partition->start;

    bamm = partition->bam - partition->start;
    fraction = (start - partition->start) & 0xfff;

back:
    itt = start - fraction;
    b = data ? 128 : 2;
    if (data && fraction > b) b = fraction;
    do {
        bam = buffer_read(itt + bamm);
        if (!bam->fault) {
            for (; b < 4096; b++) {
                unsigned char f = bam->block.data[b >> 3];
                if (f == 0) {
                    b |= 7;
                    continue;
                }
                if ((f << (b & 7)) & 0x80) goto ok;
            }
        }
        b = data ? 128 : 2;
        buffer_done(bam);
        itt += 4096;
    } while (itt < partition->end);

    itt = start - fraction;
    do {
        bam = buffer_read(itt + bamm);
        if (!bam->fault) {
            for (b = data ? 128 : 2; b < 4096; b++) {
                unsigned char f = bam->block.data[b >> 3];
                if (f == 0) {
                    b |= 7;
                    continue;
                }
                if ((f << (b & 7)) & 0x80) goto ok;
            }
        }
        buffer_done(bam);
        if (itt < 4096) break;
        itt -= 4096;
    } while (itt > partition->start);
    if (data) {
        data = 0;
        goto back;
    }
    return -ENOSPC;
ok:
    bam->block.data[b >> 3] &= ~(128 >> (b & 7));
    bam->dirty = 1;
    buffer_done(bam);
    if (partition->free) partition->free--;

    *st = itt + b;
    return 0;
}

static void free_this_sector(Partition *partition, Offset thiss) {
    Offset itt;
    unsigned int b;
    Buffer *bam;

    if (thiss < partition->start) return;
    if (thiss >= partition->end) return;

    b = (thiss - partition->start) & 0xfff;

    if (b < 2) return;

    itt = thiss - b;
    itt += partition->bam - partition->start;

    bam = buffer_read(itt);
    if (!bam->fault) {
        if (~bam->block.data[b >> 3] & (128 >> (b & 7))) {
            bam->block.data[b >> 3] |= 128 >> (b & 7);
            bam->dirty = 1;
            partition->free++;
        }
    }
    buffer_done(bam);
}

static unsigned int get_entry_size(const unsigned char *entry) {
    unsigned int size = entry[0x10] + (entry[0x11] << 8) + (entry[0x12] << 16);
    return ((entry[0x18] & FTYPE_MASK) == FTYPE_REL) ? size : (size + (entry[0x13] << 24));
}

static void set_entry_size(unsigned char *entry, unsigned int size) {
    entry[0x10] = size;
    entry[0x11] = size >> 8;
    entry[0x12] = size >> 16;
    if ((entry[0x18] & FTYPE_MASK) != FTYPE_REL) entry[0x13] = size >> 24;
}

static void fill_in_resolve(Resolve *at, Offset itt, int i, const unsigned char *entry) {
    unsigned char attr = entry[0x18];
    at->dir_sector = itt;
    at->file_offset = i;
    at->size = get_entry_size(entry);
    memcpy(at->date, entry + 0x1c, 4);
    at->attr = (((attr >> 3) & 7) * 0111) & 0755;
    switch (attr & FTYPE_MASK) {
    case FTYPE_DIR: at->attr |= S_IFDIR; at->size = 512; break;
    case FTYPE_LNK: at->attr |= S_IFLNK; break;
    default: at->attr |= S_IFREG;
    }
    at->root = offset_from(entry + 0x14);
}

static Offset get_dir_next(const Buffer *buffer) {
    unsigned char next_sector[4], c;
    const unsigned char *entry = buffer->block.data + 0x200;
    int i;

    for (i = 0; i < 4; i++) {
        entry -= 0x80;
        c = (entry[0x14] << 2) & 0xc0;
        c |= entry[0x34] & 0x30;
        c |= (entry[0x54] >> 2) & 0x0c;
        next_sector[i] = c | ((entry[0x74] >> 4) & 0x03);
    }
    return offset_from(next_sector);
}

static void set_dir_next(Buffer *buffer, Offset ezt) {
    unsigned char next_sector[4];
    unsigned char *entry = buffer->block.data + 0x200;
    int i, j;

    if (ezt == null_sector) {
        memset(next_sector, 0, 4);
    } else {
        offset_to(next_sector, ezt);
    }
    for (j = 0; j < 4; j++) {
        unsigned int c = next_sector[j] << 4;
        for (i = 0; i < 4; i++) {
            entry -= 0x20;
            entry[0x14] = (entry[0x14] & ~0x30) | (c & 0x30);
            c >>= 2;
        }
    }
    buffer->dirty = 1;
}

static int create_empty(Partition *partition, Offset *start) {
    Buffer *sbuff;
    int ret = find_free_sector(partition, start, 0);
    if (ret) return ret;
    sbuff = buffer_write(*start);
    memset(sbuff->block.data, 0, 512);
    ret = sbuff->fault;
    buffer_done(sbuff);
    return ret;
}

static int extend_dir(Resolve *at) {
    Offset empty;
    Buffer *sbuff;
    int ret;

    if (at->sectors >= 64) return -ENOSPC;

    empty = at->end_of_this_dir;
    ret = create_empty(at->partition, &empty);
    ret = buffers_sync(ret);
    if (ret) return ret;
    at->empty_sector = empty;
    at->empty_offset = 0;

    sbuff = buffer_read(at->end_of_this_dir);
    ret = sbuff->fault;
    if (!ret) {
        set_dir_next(sbuff, empty);
        ret = buffer_flush(sbuff);
    }
    buffer_done(sbuff);
    return ret;
}

static Offset get_treelink(const unsigned char *buffer) {
    unsigned char next_sector[4], c;
    const unsigned char *entry = buffer + 64;
    int i;

    for (i = 0; i < 4; i++) {
        entry -= 16;
        c = (entry[0] << 2) & 0xc0;
        c |= entry[4] & 0x30;
        c |= (entry[8] >> 2) & 0x0c;
        next_sector[i] = c | ((entry[12] >> 4) & 0x03);
    }
    return offset_from(next_sector);
}

static void set_treelink(unsigned char *buffer, Offset ezt) {
    unsigned char next_sector[4];
    unsigned char *entry = buffer + 64;
    int i, j;

    offset_to(next_sector, ezt);

    for (j = 0; j < 4; j++) {
        unsigned int c = next_sector[j] << 4;
        for (i = 0; i < 4; i++) {
            entry -= 4;
            entry[0] = (entry[0] & ~0x30) | (c & 0x30);
            c >>= 2;
        }
    }
}

static int is_valid_name_type(const Petscii *name, Petscii *type) {
    int i;
    const char *ftype;

    for (i = 0; i < 16; i++) {
        Petscii c = name[i];
        if (c == '*' || c == '?' || c == ':' || c == '=' || c == ',') return -EINVAL;
        if (c == 0) break;
    }
    if (i == 0) return -EINVAL;

    for (i = 0; i < 3; i++) {
        Petscii c = type[i];
        if (c == '*' || c == '?' || c == ':' || c == '=' || c == ',' || c == '<' || c == '>' || c == ' ') return -EINVAL;
        if (c == 0) break;
    }
    if (i == 1) {
        switch (type[0]) {
        case 'P': ftype = "PRG"; break;
        case 'S': ftype = "SEQ"; break;
        case 'U': ftype = "USR"; break;
        default: return -EINVAL;
        }
        memcpy(type, ftype, 3);
    }
    return 0;
}

static void set_name_type(unsigned char *entry, const Petscii *name, const Petscii *type) {
    int i, j;
    for (j = i = 0; i < 16; i++) {
        entry[i] = name[j];
        if (name[j]) j++;
    }
    entry += 0x19;
    for (j = i = 0; i < 3; i++) {
        entry[i] = type[j];
        if (type[j]) j++;
    }
}

typedef enum Resolve_type {
    RESOLVE_DIR, RESOLVE_FILE, RESOLVE_LINK, RESOLVE_ALL
} Resolve_type;

static int resolve_path(Resolve *at, const char *path, Resolve_type smode) {
    Offset itt, last;
    const char *p, *p2;
    int i, j, k;
    Resolve_type cmode;
    int vesszo;
    Petscii name[16 + 1];
    Petscii type[3 + 1];

    p = path;
    at->date[0] = 0;
    at->date[1] = 64;
    at->date[2] = 0;
    at->date[3] = 1;
    at->root = partition_table;
    at->sectors = 0;
    at->empty_sector = null_sector;
    at->partition = NULL;
    at->attr = S_IFDIR | 0555;
    at->size = 512;
    at->file_offset = -1;
    while (*p == '/') p++;
    if (!*p) return 0;

    if (p[0] < '0' || p[0] > '9') return 1;
    if (p[1] < '0' || p[1] > '9') return 1;
    if (p[2] != ' ') return 1;
    i = p[0] * 10 + p[1] - 529;
    if (i < 0 || i >= PARTITION_COUNT) return 1;
    if (!partitions[i].valid) return 1;
    itt = partitions[i].rootdir;
    at->root = itt;
    at->attr = S_IFDIR | 0755;
    while (*p != '/' && *p) p++;
    if (!*p) return 0;
    at->partition = &partitions[i];
    at->type[0] = at->name[0] = 0;

nextdir:
    at->sectors = 0;
    last = null_sector;
    at->empty_sector = null_sector;
    at->this_dir = itt;
    while (*p == '/') p++;
    p2 = p;
    while (*p2 != '/' && *p2) p2++;
    cmode = *p2 ? RESOLVE_DIR : smode;

    j = 0;
    while (p[j] && p[j] != '/' && p[j] != ',') j++;
    if (petscii_from_ascii(name, p, j, 16)) return 1;
    vesszo = (p[j] == ',');
    if (vesszo) {
        size_t j2 = ++j;
        while (p[j] && p[j] != '/') j++;
        j2 = j - j2;
        if (petscii_from_ascii(type, p + j - j2, j2, 3)) return 1;
        if (type[0] && !type[1]) {
            const char *t;
            switch (type[0]) {
            case 'P': t = "PRG"; break;
            case 'S': t = "SEQ"; break;
            case 'U': t = "USR"; break;
            default: return 1;
            }
            memcpy(type, t, sizeof(type));
        }
    } else memcpy(type, cmode == RESOLVE_DIR ? "DIR" : cmode == RESOLVE_LINK ? "LNK" : "PRG", sizeof(type));

    while (itt != null_sector) {
        Buffer *sbuff = buffer_read(itt);
        if (sbuff->fault) {
            int ret = sbuff->fault;
            buffer_done(sbuff);
            return ret;
        }
        at->sectors++;

        for (i = 0; i < 512; i += 32) {
            unsigned char *entry = sbuff->block.data + i;
            unsigned char attr = entry[0x18], ftype = (attr & FTYPE_MASK);
            if ((attr & 0x80) == 0 || ftype == FTYPE_DEL) {
                if ((attr & 0x87) == 0 && at->empty_sector == null_sector) {
                    at->empty_offset = i;
                    at->empty_sector = itt;
                }
                continue;
            }
            switch (cmode) {
            case RESOLVE_LINK:
            case RESOLVE_FILE: if (ftype != FTYPE_PRG && ftype != FTYPE_REL && ftype != FTYPE_LNK) continue; else break;
            case RESOLVE_DIR: if (ftype != FTYPE_DIR) continue; else break;
            default: break;
            }
            for (k = 0; k < 16; k++) {
                if (!name[k] || name[k] != entry[k]) break;
            }
            if (k < 16 && name[k] != entry[k]) continue;

            if (vesszo || cmode != RESOLVE_ALL || (ftype != FTYPE_DIR && ftype != FTYPE_LNK)) {
                unsigned char *etype = entry + 0x19;
                for (k = 0; k < 3; k++) {
                    if (!type[k] || type[k] != etype[k]) break;
                }
                if (k < 3 && type[k] != etype[k]) continue;
            }
            
            if (p[j] == '/') {
                p += j;
                itt = offset_from(entry + 0x14);
                buffer_done(sbuff);
                goto nextdir;
            }
            if (!p[j]) {
                fill_in_resolve(at, itt, i, entry);
                memcpy(at->name, name, sizeof(at->name));
                memcpy(at->type, vesszo ? type : (entry + 0x19), sizeof(at->type));
                at->type[3] = 0;
                at->end_of_this_dir = last;
                if (cmode != RESOLVE_ALL || ftype == FTYPE_DIR) {
                    buffer_done(sbuff);
                    return 0;
                }
                cmode = RESOLVE_DIR;
            }
        }
        last = itt;
        itt = get_dir_next(sbuff);
        buffer_done(sbuff);
    }
    if (at->name[0]) return 0;
    memcpy(at->name, name, sizeof(name));
    memcpy(at->type, type, sizeof(type));
    at->end_of_this_dir = last;
    return 1;
}

static int cfs011_fgetattr(const char *path, struct stat *stbuf, struct fuse_file_info *fi) {
    Resolve at;
    int ret = 0;

    *stbuf = stat_default;
    stbuf->st_uid = fuse_get_context()->uid;
    stbuf->st_gid = fuse_get_context()->gid;

    pthread_mutex_lock(&big_lock);
    if (fi == NULL) {
        ret = resolve_path(&at, path, RESOLVE_ALL);
        if (ret > 0) ret = -ENOENT;
    } else {
        File *file = (File *) (uintptr_t) fi->fh;
        Buffer *sbuff = buffer_read(file->dir_sector);
        ret = sbuff->fault;
        fill_in_resolve(&at, file->dir_sector, file->dir_offset, sbuff->block.data + file->dir_offset);
        at.partition = file->partition;
        buffer_done(sbuff);
    }
    pthread_mutex_unlock(&big_lock);

    stbuf->st_mode = at.attr;
    stbuf->st_size = at.size;
    stbuf->st_ino = at.root;
    stbuf->st_blocks = (stbuf->st_size / 512) + ((stbuf->st_size + 65535) / 512 / 128) + 1;

    if (at.partition == NULL) {
        stbuf->st_nlink = 3;
	return ret;
    }
    read_in_time(stbuf, at.date);

    stbuf->st_nlink = S_ISDIR(stbuf->st_mode) ? 3 : 1;
    return ret;
}

static int cfs011_getattr(const char *path, struct stat *stbuf) {
    return cfs011_fgetattr(path, stbuf, NULL);
}

static int cfs011_readdir(const char *path, void *buf, fuse_fill_dir_t filler,
                       off_t offset, struct fuse_file_info *fi)
{
    char name[(16 + 4) * MB_CUR_MAX + 1];
    int i, j;
    Offset itt;
    struct stat st;
    Resolve at;
    int ret;

    (void) offset;
    (void) fi;

    pthread_mutex_lock(&big_lock);
    ret = resolve_path(&at, path, RESOLVE_DIR);
    if (ret > 0) ret = -ENOENT;
    if (ret) goto fault;

    st = stat_default;
    if (at.root == partition_table) {
        st.st_mode = S_IFDIR;
        st.st_ino = partition_table;
        filler(buf, ".", &st, 0);
        filler(buf, "..", &st, 0);
        for (i = 0; i < PARTITION_COUNT; i++) {
            Partition *partition = &partitions[i];
	    if (!partition->valid) continue; // valid?
            name[0] = (i >= 9) ? '1' : '0';
            name[1] = ((i >= 9) ? '\'' : '1') + i;
            name[2] = ' ';
            petscii_to_ascii(name + 3, partition->name, 0);
            st.st_ino = partition->rootdir;
	    filler(buf, name, &st, 0);
        }
        goto fault;
    }
    itt = at.root;
    while (itt != null_sector) {
        Buffer *sbuff = buffer_read(itt);
        if (sbuff->fault) {
            ret = sbuff->fault;
            buffer_done(sbuff);
            break;
        }
        for (i = 0; i < 512; i += 32) {
            unsigned char *entry = sbuff->block.data + i, tmp;
            switch (entry[0x18] & (FTYPE_MASK | 0x80)) {
            case 0x80 | FTYPE_PRG:
            case 0x80 | FTYPE_REL: st.st_mode = S_IFREG; break;
            case 0x80 | FTYPE_DIR: st.st_mode = S_IFDIR; break;
            case 0x80 | FTYPE_LNK: st.st_mode = S_IFLNK; break;
            case FTYPE_DIR:
                st.st_mode = S_IFDIR;
                st.st_ino = offset_from(entry + 0x10);
                filler(buf, ".", &st, 0);
                st.st_ino = (at.partition == NULL) ? partition_table : offset_from(entry + 0x14);
                filler(buf, "..", &st, 0);
                continue;
            default: continue;
            }
            tmp = entry[16]; entry[16] = 0;
            j = petscii_to_ascii(name, entry, 0);
            entry[16] = tmp;
            if (st.st_mode != S_IFDIR) {
                name[j++] = ',';
                tmp = entry[0x1c]; entry[0x1c] = 0;
                petscii_to_ascii(name + j, entry + 0x19, 0);
                entry[0x1c] = tmp;
            }
            st.st_ino = offset_from(entry + 0x14);
            filler(buf, name, &st, 0);
        }
        itt = get_dir_next(sbuff);
        buffer_done(sbuff);
    }
fault:
    pthread_mutex_unlock(&big_lock);
    return ret;
}

static void recurs_truncate(Partition *partition, Offset itt, int level, unsigned int pos, unsigned int size, int modify) {
    Buffer *sbuff;
    int i, j;
    int freep = 0;
    Offset to_free[128 + 8], clean;

    if (level > 8) return;

    clean = null_sector;
    sbuff = buffer_read(itt);
    for (i = 0; i < 128; i++, pos += 512) {
        unsigned char *p = sbuff->block.data + i * 4;
        Offset itt2;
        if (pos >= size) {
            if (modify) {
                if ((p[0] & 0x40) | p[1] | p[2] | p[3]) {
                    itt2 = offset_from(p);
                    p[0] &= 0xb0;
                    p[1] = p[2] = p[3] = 0;
                    sbuff->dirty = 1;
                } else continue;
            } else itt2 = offset_from(p);
            if (itt2 != null_sector) to_free[freep++] = itt2;
            continue;
        }
        if (pos == (size & ~511)) clean = offset_from(p);
    }

    pos = (pos - 0xe000) << 3;
    for (i = 0; i < 512; i += 64, pos += 0x10000) {
        unsigned char *p = sbuff->block.data + (((i & 0x100) >> 2) + (i & 0x80) + ((i & 0x40) << 2));
        Offset itt2;
        if (pos >= size) {
            for (j = 0; j < 16; j++) if (p[j * 4] & 0x30) break;
            if (j < 16) {
                itt2 = get_treelink(p);
                for (; j < 16; j++) p[j * 4] &= ~0x30;
                sbuff->dirty = 1;
            } else continue;
            if (itt2 == null_sector) continue;
            buffer_done(sbuff);
            to_free[freep++] = itt2;
            recurs_truncate(partition, itt2, level + 1, pos, size, 0);
            sbuff = buffer_read(itt);
            continue;
        }
        if (pos == (size & ~0xffff)) {
            itt2 = get_treelink(p);
            if (itt2 == null_sector) continue;
            buffer_done(sbuff);
            recurs_truncate(partition, itt2, level + 1, pos, size, 1);
            sbuff = buffer_read(itt);
        }
    }
    buffer_done(sbuff);

    for (i = 0; i < freep; i++) free_this_sector(partition, to_free[i]);

    if (clean != null_sector) {
        sbuff = buffer_read(clean);
        for (i = (size & 511); i < 512; i++) sbuff->block.data[(i >> 7) + ((i & 0x7f) << 2)] = 0;
        sbuff->dirty = 1;
        buffer_done(sbuff);
    }
}

static int cfs011_access(const char *path, int mask) {
    Resolve at;
    int ret;

    pthread_mutex_lock(&big_lock);
    ret = resolve_path(&at, path, RESOLVE_ALL);
    pthread_mutex_unlock(&big_lock);
    if (ret > 0) return -ENOENT;
    if (ret) goto fault;

    if ((mask & X_OK) != 0 && (at.attr & 0100) == 0) ret = -EACCES;
fault:
    return ret;
}

static int cfs011_open(const char *path, struct fuse_file_info *fi) {
    Resolve at;
    int ret;

    pthread_mutex_lock(&big_lock);
    ret = resolve_path(&at, path, RESOLVE_FILE);
    if (ret > 0) ret = -ENOENT;
    if (ret) goto fault;

    if (at.partition == NULL) {
        ret = -ENOENT; goto fault;
    }
    if (fi->flags & O_TRUNC) {
        Buffer *sbuff = buffer_read(at.dir_sector);
        unsigned char *entry = sbuff->block.data + at.file_offset;
        entry = sbuff->block.data + at.file_offset;
        if (get_entry_size(entry) != 0) {
            switch (entry[0x18] & FTYPE_MASK) {
            case FTYPE_REL:
            case FTYPE_PRG:
                recurs_truncate(at.partition, at.root, 0, 0, 0, 1);
                break;
            }
            set_entry_size(entry, 0);
            sbuff->dirty = 1;
        }
        buffer_done(sbuff);
    }
    fi->fh = (uint64_t) (uintptr_t) file_open(&at);
    if ((void *)(uintptr_t)fi->fh == NULL) ret = -ENOMEM;
fault:
    pthread_mutex_unlock(&big_lock);
    return ret;
}

static int cfs011_create(const char *path, mode_t mode, struct fuse_file_info *fi) {
    Offset empty;
    Buffer *sbuff;
    Resolve at;
    int ret;

    pthread_mutex_lock(&big_lock);
    ret = resolve_path(&at, path, RESOLVE_FILE);
    if (ret < 0) goto fault;
    if (!ret) {
        if (mode & O_EXCL) {
            ret = -EEXIST; goto fault;
        }
        if (mode & O_TRUNC) {
            ret = -EEXIST; goto fault;
        }

        if (fi != NULL) {
            fi->fh = (uint64_t) (uintptr_t) file_open(&at);
            if ((void *)(uintptr_t)fi->fh == NULL) ret = -ENOMEM;
        }
        goto fault;
    }

    if (at.partition == NULL) {
        ret = -EACCES; goto fault;
    }
    if (at.this_dir == at.partition->deldir) {
        ret = -EACCES; goto fault;
    }

    ret = is_valid_name_type(at.name, at.type);
    if (ret) goto fault;

    if (!memcmp(at.type, "DEL", 3) || !memcmp(at.type, "REL", 3) ||
            !memcmp(at.type, "DIR", 3) || !memcmp(at.type, "LNK", 3)) {
        ret = -EINVAL; goto fault;
    }

    if (at.empty_sector == null_sector) {
        ret = extend_dir(&at);
        if (ret) goto fault;
    }

    empty = at.empty_sector;
    ret = create_empty(at.partition, &empty);
    if (ret) goto fault;

    sbuff = buffer_read(at.empty_sector);
    ret = sbuff->fault;
    if (!ret) {
        unsigned char *entry = sbuff->block.data + at.empty_offset;
        set_name_type(entry, at.name, at.type);
        memset(entry + 0x10, 0, 4);
        entry[0x14] &= ~0x80;
        offset_insert(entry + 0x14, empty);
        entry[0x18] = 0xc0 | FTYPE_PRG | ((mode & 0700) >> 3) | (mode & 070) | ((mode & 07) << 3);
        fill_in_time(entry + 0x1c, time(NULL));
        sbuff->dirty = 1;
    }
    buffer_done(sbuff);
    if (ret) goto fault;
    if (fi != NULL) {
        at.root = empty;
        at.dir_sector = at.empty_sector;
        at.file_offset = at.empty_offset;
        fi->fh = (uint64_t) (uintptr_t) file_open(&at);
        if ((void *)(uintptr_t)fi->fh == NULL) ret = -ENOMEM;
    }
fault:
    pthread_mutex_unlock(&big_lock);
    return ret;
}

static int cfs011_mknod(const char *path, mode_t mode, dev_t rdev) {
    (void) rdev;
    int ret;

    if (!S_ISREG(mode)) {
        return -ENOSYS;
    }

    ret = cfs011_create(path, mode | O_EXCL, NULL);
    pthread_mutex_lock(&big_lock);
    ret = buffers_sync(ret);
    pthread_mutex_unlock(&big_lock);
    return ret;
}

static Offset read_tree_search(Offset itt, off_t offset, int *ret) {
    off_t m = offset / 512 / 128;
    off_t m2 = 0;
    unsigned int level = 0;

    while (m) {
        m--;
        m2 <<= 1; m2 |= m & 1; m >>= 1;
        m2 <<= 1; m2 |= m & 1; m >>= 1;
        m2 <<= 1; m2 |= m & 1; m >>= 1;
        level++;
    }

    while (level && itt != null_sector) {
        Buffer *tbuff = buffer_read(itt);
        if (tbuff->fault) {
            *ret = tbuff->fault;
            itt = null_sector;
        } else {
            itt = get_treelink(tbuff->block.data + (m2 & 7) * 64);
            m2 >>= 3;
            level--;
        }
        buffer_done(tbuff);
    }
    return itt;
}

static int cfs011_read(const char *path, char *buf, size_t size, off_t offset,
                    struct fuse_file_info *fi)
{
    off_t last_offset = 1;
    off_t fsize;
    Offset root, itt, itt2;
    Buffer *sbuff, *tbuff;
    unsigned char *entry;
    int read, ret;
    File *file;
    (void) path;

    pthread_mutex_lock(&big_lock);

    read = 0;
    file = (File *) (uintptr_t) fi->fh;

    sbuff = buffer_read(file->dir_sector);
    ret = sbuff->fault;
    entry = sbuff->block.data + file->dir_offset;
    fsize = get_entry_size(entry);

    fsize = (fsize > offset) ? (fsize - offset) : 0;
    if (size > (size_t)fsize) size = fsize;

    root = offset_from(entry + 0x14);
    buffer_done(sbuff);

    goto start;
    while (size) {
        size_t size2;

        if (last_offset != (offset & ~0xffff)) {
        start:
            last_offset = offset & ~0xffff;
            itt = read_tree_search(root, offset, &ret);
        }
   
        size2 = size;
        if (size2 > (512u - (offset & 511u))) {
            size2 = 512u - (offset & 511u);
        }

        itt2 = itt;
        if (itt2 != null_sector) {
            tbuff = buffer_read(itt);
            if (tbuff->fault) {
                ret = tbuff->fault;
                itt2 = null_sector;
            } else {
                itt2 = offset_from(tbuff->block.data + ((offset >> 7) & 0x1fc));
            }
            buffer_done(tbuff);
        }

        if (itt2 == null_sector) {
            memset(buf, 0, size2);
        } else {
            int i;
            Buffer *b = buffer_read_once(itt2);
            const unsigned char *data = b->block.data;
            if (size2 == 512) {
                for (i = 0; i < 128; i++) {
                    buf[i] = data[i * 4];
                    buf[i + 128] = data[i * 4 + 1];
                    buf[i + 256] = data[i * 4 + 2];
                    buf[i + 384] = data[i * 4 + 3];
                }
            } else {
                unsigned int j;
                i = (offset & 511);
                for (j = 0; j < size2; j++, i++) buf[j] = data[(i >> 7) + ((i & 0x7f) << 2)];
            }
            if (b->fault) ret = b->fault;
            buffer_done(b);
        }

        buf += size2;
        read += size2;
        offset += size2;
        size -= size2;
    }
    ret = read;
    pthread_mutex_unlock(&big_lock);
    return ret;
}

static int cfs011_write(const char *path, const char *buf, size_t size,
                     off_t offset, struct fuse_file_info *fi)
{
    off_t last_offset = 1;
    Offset itt = null_sector, empty;
    Buffer *sbuff, *tbuff;
    unsigned char *entry;
    int write, ret = 0, ret2;
    File *file;
    (void) path;

    pthread_mutex_lock(&big_lock);

    write = 0;
    file = (File *) (uintptr_t) fi->fh;

    while (size) {
        unsigned int i, j;
        size_t size2 = size;
        int holecheck;

        if (size2 > (512u - (offset & 511u))) {
            size2 = 512u - (offset & 511u);
        }
        holecheck = (size2 == 512);

        if (last_offset != (offset & ~0xffff)) {
            off_t m = offset / 512 / 128;
            off_t m2 = 0;
            unsigned int level = 0;

            while (m) {
                m--;
                m2 <<= 1; m2 |= m & 1; m >>= 1;
                m2 <<= 1; m2 |= m & 1; m >>= 1;
                m2 <<= 1; m2 |= m & 1; m >>= 1;
                level++;
            }

            itt = file->treeroot;

            while (level) {
                Offset itt2;

                tbuff = buffer_read(itt);
                if (tbuff->fault) {
                    ret = tbuff->fault;
                    buffer_done(tbuff);
                    goto hole;
                }
                itt2 = get_treelink(tbuff->block.data + (m2 & 7) * 64);
                buffer_done(tbuff);
                if (itt2 == null_sector) {// it does not exists yet! (empty sector)
                    if (holecheck) {
                        for (i = 0; i < size2; i++) {
                            if (buf[i] != 0) break;
                        }
                        if (i == size2) goto hole;
                        holecheck = 0;
                    }
                    empty = file->lasttree;
                    ret2 = create_empty(file->partition, &empty);
                    if (ret2) {
                        ret = ret2;
                        goto hole;
                    }

                    tbuff = buffer_read(itt);
                    set_treelink(tbuff->block.data + (m2 & 7) * 64, empty);
                    tbuff->dirty = 1;
                    buffer_done(tbuff);
                    itt = empty;
                } else {
                    itt = itt2;
                }

                m2 >>= 3;
                level--;
            }
            last_offset = offset & ~0xffff;
        }

        tbuff = buffer_read(itt);
        empty = offset_from(tbuff->block.data + ((offset >> 7) & 0x1fc));
        buffer_done(tbuff);

        if (empty == null_sector) {
            if (holecheck) {
                for (i = 0; i < size2; i++) {
                    if (buf[i] != 0) break;
                }
                if (i == size2) goto hole;
                holecheck = 0;
            }
            empty = file->lastblock;
            ret2 = find_free_sector(file->partition, &empty, 1);
            if (ret2) {
                ret = ret2;
                goto hole;
            }

            tbuff = buffer_read(itt);
            offset_insert(tbuff->block.data + ((offset >> 7) & 0x1fc), empty);
            tbuff->dirty = 1;
            buffer_done(tbuff);

            if (size2 != 512) {
                sbuff = buffer_write_once(empty);
                memset(sbuff->block.data, 0, 512);
            }
        } else {
            if (size2 != 512) sbuff = buffer_read_once(empty);
        }
        if (size2 == 512) {
            sbuff = buffer_write_once(empty);
            for (i = 0; i < 128; i++) {
                unsigned char *data = sbuff->block.data + i * 4;
                data[0] = buf[i];
                data[1] = buf[i + 128];
                data[2] = buf[i + 256];
                data[3] = buf[i + 384];
            }
        } else {
            i = (offset & 511);
            for (j = 0; j < size2; j++, i++) sbuff->block.data[(i >> 7) + ((i & 0x7f) << 2)] = buf[j];
            sbuff->dirty = 1; 
        }
        ret2 = buffer_flush(sbuff);
        if (ret2) ret = ret2;
        file->lastblock = empty;
    hole:
        write += size2;
        buf += size2;
        offset += size2;
        size -= size2;
    }

    sbuff = buffer_read(file->dir_sector);
    entry = sbuff->block.data + file->dir_offset;

    if (get_entry_size(entry) < offset) {
        set_entry_size(entry, offset);
        sbuff->dirty = 1;
    }
    buffer_done(sbuff);
    if (ret == 0) ret = write;

    pthread_mutex_unlock(&big_lock);
    return ret;
}

static int change_attr(mode_t mode, Offset itt, unsigned int offset) {
    Buffer *sbuff = buffer_read(itt);
    int ret = sbuff->fault;

    if (!ret) {
        unsigned char *entry = sbuff->block.data + offset;
        unsigned char attr = (entry[0x18] & 0xc7) | ((mode & 0700) >> 3) | (mode & 070) | ((mode & 07) << 3);
        if (entry[0x18] != attr) {
            entry[0x18] = attr;
            sbuff->dirty = 1;
        }
        ret = (attr & 0x87) == 0x83;
    }
    buffer_done(sbuff);
    return ret;
}

static int cfs011_chmod(const char *path, mode_t mode) {
    Resolve at;
    int ret;

    pthread_mutex_lock(&big_lock);
    ret = resolve_path(&at, path, RESOLVE_ALL);
    if (ret > 0) ret = -ENOENT;
    if (ret) goto fault;

    if (at.partition == NULL) {
        ret = -EACCES; goto fault;
    }
    if (at.root == at.partition->deldir) {
        ret = -EACCES; goto fault;
    }
    if (at.root != at.partition->rootdir) {
        ret = change_attr(mode, at.dir_sector, at.file_offset);
        if (ret > 0) ret = change_attr(mode, at.root, 0);
    }
fault:
    //ret = buffers_sync(ret);
    pthread_mutex_unlock(&big_lock);
    return ret;
}

static int cfs011_chown(const char *path, uid_t uid, gid_t gid) {
    Resolve at;
    (void) uid;
    (void) gid;
    int ret;

    pthread_mutex_lock(&big_lock);
    ret = resolve_path(&at, path, RESOLVE_ALL);
    if (ret > 0) ret = -ENOENT;
    if (ret) goto fault;

    if (at.partition == NULL) {
        ret = -EACCES; goto fault;
    }
fault:
    pthread_mutex_unlock(&big_lock);
    return ret;
}

static int change_time(const unsigned char *newtime, Offset itt, unsigned int offset) {
    Buffer *sbuff = buffer_read(itt);
    int ret = sbuff->fault;

    if (!ret) {
        unsigned char *entry = sbuff->block.data + offset;
        ret = (entry[0x18] & 0x87) == 0x83;
        entry += 0x1c;
        if (memcmp(entry, newtime, 4) != 0) {
            memcpy(entry, newtime, 4);
            sbuff->dirty = 1;
        }
    }
    buffer_done(sbuff);
    return ret;
}

static int cfs011_utimens(const char *path, const struct timespec tv[2]) {
    Resolve at;
    int ret;

    pthread_mutex_lock(&big_lock);
    ret = resolve_path(&at, path, RESOLVE_ALL);
    if (ret > 0) ret = -ENOENT;
    if (ret) goto fault;

    if (at.partition == NULL) {
        ret = -EACCES; goto fault;
    }
    if (at.root == at.partition->deldir) {
        ret = -EACCES; goto fault;
    }
    if (at.root != at.partition->rootdir) {
        unsigned char newtime[4];
        fill_in_time(newtime, tv[1].tv_sec);
        ret = change_time(newtime, at.dir_sector, at.file_offset);
        if (ret > 0) ret = change_time(newtime, at.root, 0);
    }
fault:
    //ret = buffers_sync(ret);
    pthread_mutex_unlock(&big_lock);
    return ret;
}

static int cfs011_symlink(const char *from, const char *to) {
    Offset empty;
    Resolve at;
    Buffer *sbuff;
    unsigned char *entry;
    int i;
    int ret;

    pthread_mutex_lock(&big_lock);
    ret = resolve_path(&at, to, RESOLVE_LINK);
    if (!ret) ret = -EEXIST;
    if (ret < 0) goto fault;

    if (at.partition == NULL) {
        ret = -EACCES; goto fault;
    }
    if (at.this_dir == at.partition->deldir) {
        ret = -EACCES; goto fault;
    }

    ret = is_valid_name_type(at.name, at.type);
    if (ret) goto fault;

    if (memcmp(at.type, "LNK", 3)) {
        ret = -EINVAL; goto fault;
    }

    if (at.empty_sector == null_sector) {
        ret = extend_dir(&at);
        if (ret) goto fault;
    }

    empty = at.empty_sector;
    ret = create_empty(at.partition, &empty);
    if (ret) goto fault;

    sbuff = buffer_read(empty);
    ret = sbuff->fault;
    if (ret) goto fault;
    petscii_from_ascii(sbuff->block.data, from, strlen(from), 511);
    i = strlen((const char *)sbuff->block.data);
    sbuff->dirty = 1;
    buffer_done(sbuff);

    sbuff = buffer_read(at.empty_sector);
    entry = sbuff->block.data + at.empty_offset;
    set_name_type(entry, at.name, (const Petscii *)"LNK");
    entry[0x10] = i;
    entry[0x11] = i >> 8;
    entry[0x12] = 0;
    entry[0x13] = 0;
    entry[0x14] &= ~0x80;
    offset_insert(entry + 0x14, empty);
    entry[0x18] = 0xf8 | FTYPE_LNK;
    fill_in_time(entry + 0x1c, time(NULL));
    sbuff->dirty = 1;
    buffer_done(sbuff);
fault:
    ret = buffers_sync(ret);
    pthread_mutex_unlock(&big_lock);
    return ret;
}

static int cfs011_rename(const char *from, const char *to) {
    Resolve at, at2;
    Buffer *sbuff;
    unsigned char *entry;
    int ret;

    pthread_mutex_lock(&big_lock);
    ret = resolve_path(&at2, from, RESOLVE_ALL);
    if (ret > 0) ret = -ENOENT;
    if (ret) goto fault;

    if (at2.partition == NULL) {
        ret = -EACCES; goto fault;
    }
    if (at2.root == at2.partition->deldir) {
        ret = -EACCES; goto fault;
    }
    ret = resolve_path(&at, to, S_ISDIR(at2.attr) ? RESOLVE_DIR : S_ISLNK(at2.attr) ? RESOLVE_LINK : RESOLVE_FILE);
    if (ret < 0) goto fault;
    if (!ret) {
        if (at.partition != at2.partition || at.dir_sector != at2.dir_sector || at.file_offset != at2.file_offset) ret = -EEXIST; 
        goto fault;
    }
    if (at.this_dir == at.partition->deldir) {
        ret = -EACCES; goto fault;
    }
    if (at.partition != at2.partition) {
        ret = -EXDEV; goto fault;
    }
    if (at.this_dir != at2.this_dir) {
        ret = -EXDEV; goto fault;
    }

    ret = is_valid_name_type(at.name, at.type);
    if (ret) goto fault;

    if (memcmp(at.type, at2.type, 3) && (!memcmp(at.type, "DEL", 3) || !memcmp(at.type, "REL", 3) ||
        !memcmp(at.type, "DIR", 3) || !memcmp(at.type, "LNK", 3))) {
        ret = -EINVAL; goto fault;
    }

    sbuff = buffer_read(at2.dir_sector);
    at.empty_sector = at2.dir_sector;
    at.empty_offset = at2.file_offset;

    entry = sbuff->block.data + at.empty_offset;
    set_name_type(entry, at.name, (entry[0x18] & FTYPE_MASK) == FTYPE_PRG ? at.type : at2.type);
    sbuff->dirty = 1;
    buffer_done(sbuff);
fault:
    ret = buffers_sync(ret);
    pthread_mutex_unlock(&big_lock);
    return ret;
}

static int cfs011_mkdir(const char *path, mode_t mode) {
    Offset empty;
    Resolve at;
    Buffer *sbuff;
    unsigned char *entry;
    int i, j;
    time_t ee;
    int ret;

    pthread_mutex_lock(&big_lock);
    ret = resolve_path(&at, path, RESOLVE_DIR);
    if (!ret) ret = -EEXIST;
    if (ret < 0) goto fault;

    if (at.partition == NULL) {
        ret = -EACCES; goto fault;
    }
    if (at.this_dir == at.partition->deldir) {
        ret = -EACCES; goto fault;
    }

    ret = is_valid_name_type(at.name, at.type);
    if (ret) goto fault;

    if (memcmp(at.type, "DIR", 3)) {
        ret = -EINVAL; goto fault;
    }

    if (at.empty_sector == null_sector) {
        ret = extend_dir(&at);
        if (ret) goto fault;
    }

    empty = at.empty_sector;
    ret = find_free_sector(at.partition, &empty, 0);
    if (ret) goto fault;

    sbuff = buffer_write(empty);
    entry = sbuff->block.data;
    memset(entry, 0, 512);
    for (j = i = 0; i < 16; i++) {
        if (at.name[j]) entry[i] = at.name[j++];
        else entry[i] = 0x20;
    }
    offset_to(entry + 0x10, empty);
    offset_to(entry + 0x14, at.this_dir);
    entry[0x18] = 0x40 | FTYPE_DIR | ((mode & 0700) >> 3) | (mode & 070) | ((mode & 07) << 3);
    memcpy(entry + 0x19, "DIR", 3);
    ee = time(NULL);
    fill_in_time(entry + 0x1c, ee);
    ret = buffer_flush(sbuff);
    buffer_done(sbuff);
    ret = buffers_sync(ret);
    if (ret) goto fault;

    sbuff = buffer_read(at.empty_sector);
    ret = sbuff->fault;
    if (ret) goto fault;
    sbuff->dirty = 1;
    entry = sbuff->block.data + at.empty_offset;
    set_name_type(entry, at.name, (const Petscii *)"DIR");
    memset(entry + 0x10, 0, 4);
    entry[0x14] &= ~0x80;
    offset_insert(entry + 0x14, empty);
    entry[0x18] = 0xc0 | FTYPE_DIR | ((mode & 0700) >> 3) | (mode & 070) | ((mode & 07) << 3);
    fill_in_time(entry + 0x1c, ee);
    ret = buffer_flush(sbuff);
    buffer_done(sbuff);
fault:
    pthread_mutex_unlock(&big_lock);
    return ret;
}

static int cfs011_ftruncate(const char *path, off_t size, struct fuse_file_info *fi) {
    Resolve at;
    Buffer *sbuff;
    unsigned char *entry;
    unsigned int size2;
    int ret = 0;

    pthread_mutex_lock(&big_lock);

    if (fi == NULL) {
        ret = resolve_path(&at, path, RESOLVE_FILE);
        if (ret > 0) ret = -ENOENT;
        if (ret) goto fault;

        if (at.partition == NULL) {
            ret = -ENOENT; goto fault;
        }
    } else {
        File *file = (File *) (uintptr_t) fi->fh;
        at.dir_sector = file->dir_sector;
        at.file_offset = file->dir_offset;
        at.partition = file->partition;
        at.root = file->treeroot;
    }

    sbuff = buffer_read(at.dir_sector);
    entry = sbuff->block.data + at.file_offset;
    size2 = get_entry_size(entry);

    switch (entry[0x18] & FTYPE_MASK) {
    case FTYPE_REL:
        if (size > 0xffffff) {
            ret = -EFBIG; goto fault;
        }
        /* fall through */
    case FTYPE_PRG:
        if (size <= size2) {
            recurs_truncate(at.partition, at.root, 0, 0, size, size < 65536);
        }
        break;
    case FTYPE_LNK:
        if (size > 0x200) {
            ret = -EFBIG; goto fault;
        }
        break;
    }

    if (size != size2) {
        set_entry_size(entry, size);
        sbuff->dirty = 1;
    }
    buffer_done(sbuff);
fault:
    pthread_mutex_unlock(&big_lock);
    return ret;
}

static int cfs011_truncate(const char *path, off_t size) {
    int ret = cfs011_ftruncate(path, size, NULL);
    pthread_mutex_lock(&big_lock);
    ret = buffers_sync(ret);
    pthread_mutex_unlock(&big_lock);
    return ret;
}

static int is_dir_empty(const Buffer *buffer, int except) {
    int i;
    for (i = 0; i < 512; i += 32) {
        const unsigned char *entry = buffer->block.data + i;
        unsigned char attr = entry[0x18];
        if ((attr & 0x80) != 0x00 && (attr & 0x87) != 0x03) {
            if (i != except) return 0;
        }
    }
    return 1;
}

static int delete_entry(const Resolve *at) {
    int ret;
    Buffer *sbuff = buffer_read(at->dir_sector);
    ret = sbuff->fault;
    if (!ret) ret = sbuff->block.data[at->file_offset + 0x18] & 0x87;

    if (ret >= 0 && at->end_of_this_dir != null_sector) {
        if (is_dir_empty(sbuff, at->file_offset)) {
            Offset itt = get_dir_next(sbuff);
            buffer_done(sbuff);
            sbuff = buffer_read(at->end_of_this_dir);
            if (sbuff->fault) {
                ret = sbuff->fault;
                buffer_done(sbuff);
                return ret;
            } 
            set_dir_next(sbuff, itt);
            buffer_done(sbuff);
            free_this_sector(at->partition, at->dir_sector);
            return ret;
        }
    }
    if (ret >= 0) {
        sbuff->block.data[at->file_offset + 0x18] &= 0x78;
        sbuff->dirty = 1;
    }
    buffer_done(sbuff);
    return ret;
}

static int cfs011_unlink(const char *path) {
    Resolve at;
    int ret;

    pthread_mutex_lock(&big_lock);
    ret = resolve_path(&at, path, RESOLVE_FILE);
    if (ret > 0) ret = -ENOENT;
    if (ret) goto fault;

    if (at.partition == NULL) {
        ret = -ENOENT; goto fault;
    }
    if (file_lookup(&at) != NULL) {
        ret = -EBUSY; goto fault;
    }

    ret = delete_entry(&at);
    if (ret < 0) goto fault;

    switch (ret) {
    case 0x80: break;
    case 0x81:
    case 0x82:
        recurs_truncate(at.partition, at.root, 0, 0, 0, 0);
        free_this_sector(at.partition, at.root);
        break;
    case 0x84: free_this_sector(at.partition, at.root); break;
    }
    ret = 0;
fault:
    ret = buffers_sync(ret);
    pthread_mutex_unlock(&big_lock);
    return ret;
}

static int rmdir_walk(const Resolve *at, int delete) {
    int ret = 0;
    Offset itt, next;

    for (itt = at->root; itt != null_sector; itt = next) {
        Buffer *sbuff = buffer_read(itt);
        ret = sbuff->fault;
        if (ret) {
            buffer_done(sbuff);
            break;
        }
        if (!delete && !is_dir_empty(sbuff, -1)) {
            buffer_done(sbuff);
            return -ENOTEMPTY;
        }
        next = get_dir_next(sbuff);
        buffer_done(sbuff);
        if (delete) free_this_sector(at->partition, itt);
    }
    return ret;
}

static int cfs011_rmdir(const char *path) {
    Resolve at;
    int ret;

    pthread_mutex_lock(&big_lock);
    ret = resolve_path(&at, path, RESOLVE_DIR);
    if (ret > 0) ret = -ENOENT;
    if (ret) goto fault;

    if (at.partition == NULL) {
        ret = -EACCES; goto fault;
    }
    if (at.root == at.partition->deldir) {
        ret = -EACCES; goto fault;
    }

    ret = rmdir_walk(&at, 0);
    if (ret) goto fault;

    ret = delete_entry(&at);
    if (ret < 0) goto fault;

    ret = rmdir_walk(&at, 1);
fault:
    ret = buffers_sync(ret);
    pthread_mutex_unlock(&big_lock);
    return ret;
}

static int cfs011_statfs(const char *path, struct statvfs *stbuf) {
    Resolve at;
    static int fs_free_updated = 0;
    int i, j;
    Offset itt;
    int ret;

    pthread_mutex_lock(&big_lock);
    ret = resolve_path(&at, path, RESOLVE_ALL);
    if (ret > 0) ret = -ENOENT;
    if (ret) goto fault;

    if (!fs_free_updated) {
        for (i = 0; i < PARTITION_COUNT; i++) {
            off_t freeblk;
            Partition *partition = &partitions[i];
            if (!partition->valid) continue;
            partition->size = partition->end - partition->start;
            freeblk = 0;

            itt = partition->bam;
            do {
                Buffer *sbuff = buffer_read(itt);
                if (sbuff->fault) {
                    ret = sbuff->fault;
                } else {
                    unsigned char f;
                    for (j = 0; j < 512; j++) {
                        f = sbuff->block.data[j];
                        if (f == 0xff) {
                            freeblk += 8;
                            continue;
                        }
                        while (f) {
                            freeblk += f & 1;
                            f >>= 1;
                        }
                    }
                    f = sbuff->block.data[0];
                    if (f & 0x80) freeblk--;
                    if (f & 0x40) freeblk--;
                }
                buffer_done(sbuff);
                itt += 4096;
                partition->size -= 2;
            } while (itt < partition->end);
            partition->free = freeblk;
        }

        fs_free_updated = 1;
    }
    if (at.partition != NULL) {
        stbuf->f_bfree = at.partition->free;
        stbuf->f_blocks = at.partition->size;
    } else {
        stbuf->f_bfree = 0;
        stbuf->f_blocks = 0;
        for (i = 0; i < PARTITION_COUNT; i++) {
            Partition *partition = &partitions[i];
            if (!partition->valid) continue;
            if (partition->rootdir == at.root) {
                stbuf->f_bfree = partition->free;
                stbuf->f_blocks = partition->size;
                break;
            }
            stbuf->f_bfree += partition->free;
            stbuf->f_blocks += partition->size;
        }
    }
    stbuf->f_bavail = stbuf->f_bfree;
    //stbuf->f_type = 0x64;
    stbuf->f_bsize = 512;
    stbuf->f_frsize = 512;
    stbuf->f_files = 0;
    stbuf->f_ffree = 0;
#ifdef __CYGWIN__
    stbuf->f_namemax = 40;
#else
    stbuf->f_namemax = 20;
#endif
fault:
    pthread_mutex_unlock(&big_lock);
    return ret;
}

static int cfs011_fsync(const char *path, int isdatasync,
                     struct fuse_file_info *fi)
{
    (void) path;
    (void) isdatasync;
    (void) fi;
    int ret;

    pthread_mutex_lock(&big_lock);
    ret = buffers_sync(0);
    pthread_mutex_unlock(&big_lock);
    return ret;
}

static int cfs011_fsyncdir(const char *path, int isdatasync,
                     struct fuse_file_info *fi)
{
    (void) path;
    (void) isdatasync;
    (void) fi;
    int ret;

    pthread_mutex_lock(&big_lock);
    ret = buffers_sync(0);
    pthread_mutex_unlock(&big_lock);
    return ret;
}

static int cfs011_readlink(const char *path, char *out, size_t size) {
    Resolve at;
    Buffer *sbuff;
    char buf[512*MB_CUR_MAX + 1];
    int ret;
    size_t s;

    pthread_mutex_lock(&big_lock);
    ret = resolve_path(&at, path, RESOLVE_LINK);
    if (ret > 0) ret = -ENOENT;
    if (ret) goto fault;
    
    if (at.partition == NULL) {
        ret = -ENOENT; goto fault;
    }

    sbuff = buffer_read(at.root);
    ret = sbuff->fault;
    s = petscii_to_ascii(buf, sbuff->block.data, 1) + 1;
    buffer_done(sbuff);
    memcpy(out, buf, (s < size) ? s : size);
fault:
    pthread_mutex_unlock(&big_lock);
    return ret;
}

static int cfs011_flush(const char *path, struct fuse_file_info *fi) {
    (void) path;
    (void) fi;
    int ret;

    pthread_mutex_lock(&big_lock);
    ret = buffers_sync(0);
    pthread_mutex_unlock(&big_lock);
    return ret;
}

static int cfs011_release(const char *path, struct fuse_file_info *fi) {
    (void) path;
    File *file = (File *) (uintptr_t) fi->fh;

    pthread_mutex_lock(&big_lock);
    file_close(file);
    pthread_mutex_unlock(&big_lock);
    return 0;
}

static void *cfs011_init(struct fuse_conn_info *conn) {
    int i, fault;
    Buffer *sbuff;
    Offset backup_table;
    struct tm epoch;

    stat_default.st_blksize = 512;
    stat_default.st_mtime = time(NULL);
    stat_default.st_atime = stat_default.st_mtime;
    stat_default.st_ctime = stat_default.st_mtime;

    localtime_r(&stat_default.st_mtime, &epoch);
    epoch.tm_sec = 0;
    epoch.tm_min = 0;
    epoch.tm_hour = 23;
    epoch.tm_mday = 29;
    epoch.tm_mon = 10;
    epoch.tm_year = 79;
    epoch.tm_wday = 0;
    epoch.tm_yday = 0;
    epoch.tm_isdst = 0;
    epoch_time = mktime(&epoch);

#ifdef FUSE_CAP_BIG_WRITES
    conn->want |= FUSE_CAP_BIG_WRITES;
#endif
#ifdef FUSE_CAP_ATOMIC_O_TRUNC
    conn->want |= FUSE_CAP_ATOMIC_O_TRUNC;
#endif

    buffers_init();

    sbuff = buffer_read(0);
    fault = sbuff->fault || (memcmp(sbuff->block.data + 0x08, "C64 CFS V 0.11B ", 16) != 0) || offset_geometry(sbuff->block.data + 4);
    if (!fault) {
        partition_table = offset_from(sbuff->block.data + 0x18);
        backup_table = offset_from(sbuff->block.data + 0x1c);
    }
    buffer_done(sbuff);
    if (fault) return NULL;

    sbuff = buffer_read(partition_table);
    if (sbuff->fault) {
        buffer_done(sbuff);
        partition_table = backup_table;
        sbuff = buffer_read(partition_table);
    }
    if (!sbuff->fault) {
        for (i = 0; i < PARTITION_COUNT; i++) {
            unsigned char *entry = sbuff->block.data + i * 32;
            Partition *partition = &partitions[i];
            partition->valid = (entry[0x10] & 0x80) && (entry[0x14] & 0xB0) == 0x10;
            partition->rootdir = offset_from(entry + 0x1c);
            partition->start = offset_from(entry + 0x10);
            partition->end = offset_from(entry + 0x14);
            partition->deldir = offset_from(entry + 0x18);
            partition->bam = (entry[0x1c] >> 7) + partition->start;
            memcpy(partition->name, entry, 16);
            partition->name[16] = 0;
            if (partition->rootdir < partition->start || partition->rootdir >= partition->end) partition->valid = 0;
        }
    }
    buffer_done(sbuff);
    return NULL;
}

static void cfs011_destroy(void *userdata) {
    (void) userdata;
    buffers_sync(0);
}

static struct fuse_operations cfs011_oper = {
    .getattr	= cfs011_getattr,
    .access	= cfs011_access,
    .readlink	= cfs011_readlink,
    .mknod	= cfs011_mknod,
    .mkdir	= cfs011_mkdir,
    .unlink	= cfs011_unlink,
    .rmdir	= cfs011_rmdir,
    .symlink	= cfs011_symlink,
    .rename	= cfs011_rename,
    .link       = NULL,
    .chmod	= cfs011_chmod,
    .chown	= cfs011_chown,
    .truncate	= cfs011_truncate,
    .utimens	= cfs011_utimens,
    .open	= cfs011_open,
    .read	= cfs011_read,
    .write	= cfs011_write,
    .statfs	= cfs011_statfs,
    .flush      = cfs011_flush,
    .release	= cfs011_release,
    .fsync	= cfs011_fsync,
    .setxattr   = NULL,
    .getxattr   = NULL,
    .listxattr  = NULL,
    .removexattr= NULL,
    .opendir    = NULL,
    .readdir	= cfs011_readdir,
    .releasedir = NULL,
    .fsyncdir   = cfs011_fsyncdir,
    .init       = cfs011_init,
    .destroy    = cfs011_destroy,
    .create	= cfs011_create,
    .ftruncate	= cfs011_ftruncate,
    .fgetattr	= cfs011_fgetattr,
};

int main(int argc, char *argv[]) {
    const char *message;
    setlocale(LC_ALL, "");

    puts("FUSECFS011 v2.0.3 (C) 2005-2020 Kajtar Zsolt (Soci/Singular)");
    if (argc < 2) {
        printf("Usage: %s <device> <mount_point> [<FUSE library options>]\nUnmount: fusermount -u <mount_point>\n", argv[0]);
	return EXIT_FAILURE;
    }

    message = block_init(argv[1]);
    if (message != NULL) {
        printf("Cannot open image file \"%s\": %s\n", argv[1], message);
	return EXIT_FAILURE;
    }
    argv[1] = argv[0];

    return fuse_main(argc - 1, argv + 1, &cfs011_oper, NULL);
}
